import json
import sys
import requests


def isDebugMode() -> bool:
    """Return if the debugger is currently active"""
    return hasattr(sys, 'gettrace') and sys.gettrace() is not None

class RestAssetCommunicator(object):
    
    '''Constructor '''
    def __init__(self,server,bearer,version,prj_name):
       
        self.bearer=bearer 
        self.coords=server
        self.version=version
        self.prjname=prj_name
        self.info={}
        self.mapNameType={}
        self.mapNameAttribs={}
        self.referenceCache={}
        self.schemaId=""
        self.schemaStates=None
        res=self.findObjectSchemas()
        schemas=res['objectschemas']
        for ele in schemas:
            if ele['name'] == prj_name :
                self.info = ele
                self.schemaId=str(ele['id'])
        
        if self.info == {} :
            raise Exception("Project Name \""+prj_name+"\" not found.")
        
        print ("Found project id "+self.schemaId) 
        if isDebugMode(): 
            print("caching types...") 
        self.objectTypeList=self.findObjectTypesFlatList()
        if isDebugMode(): 
            print("caching attributes for types...") 
        for t in self.objectTypeList:
            self.mapNameType[t['name']]=t
        self.schemaStates=self.findStatusTypes()
        if isDebugMode(): 
            print("init done.")   
        pass

    ''' _private_get: Send as GET HTTP the query string passed as parameter
    @returns a dictionary
    ''' 
    def _private_get(self,query,params=None):
         
        if isDebugMode():
            print(f"GET:{query}")
        
        response = requests.get(query, params=params,headers={'Authorization':"Bearer "+self.bearer })

        
        return json.loads(response.content)
        
    def _private_put(self,query,payload):
        if isDebugMode():
            print(f"PUT:{query} {payload}")
        myHeaders = {
		
        "Content-Type": "application/json",
		"Accept": "Application/json",
                "Authorization": "Bearer {}".format(self.bearer) }
        j=json.dumps(payload)
        response = requests.put(query, data=j, headers=myHeaders)
        if response.status_code == 401:
            raise Exception("Client must be authenticated to access this resource")
        if response.status_code == 500:
            raise Exception("Something went wrong. Please try again later. If the error persists, contact our support team, func: "+query)
        
        if response.status_code == 404:
            raise Exception("Not found: "+query)
            
        if (response.status_code != 200):            raise Exception("Returned code "+str(response.status_code) + " request: "+query)
    
    ''' _private_delete: Send as DELETE HTTP the query string passed as parameter '''
    def _private_delete(self,query):
        if isDebugMode():
            print(f"DELETE:{query}")
        
        myHeaders = {
		
        "Content-Type": "application/json",
		"Accept": "Application/json",
                "Authorization": "Bearer {}".format(self.bearer) }
        response = requests.delete(query, headers=myHeaders)
        if response.status_code == 401:
            raise Exception("Client must be authenticated to access this resource")
        if response.status_code == 500:
            raise Exception("Something went wrong. Please try again later. If the error persists, contact our support team, func: "+query)
        
        if response.status_code == 404:
            raise Exception("Not found: "+query)
            
        if (response.status_code != 200):
            raise Exception("Returned code "+str(response.status_code) + " request: "+query)
        


    ''' _private_generic_post: Send a Post HTTP to generic coords address, with the payload expressed as a dictionary
        @returns a dictionary
    '''
    def _private_generic_post(self,coords,payload):
        if isDebugMode():
            print(f"POST:{coords} {payload}")
        
        myHeaders = {
		
        "Content-Type": "application/json",
		"Accept": "Application/json",
                "Authorization": "Bearer {}".format(self.bearer) }
        #headers = {'Authorization':"Bearer "+self.bearer }
        
        j=json.dumps(payload)
        response = requests.post(coords, data=j, headers=myHeaders)
        if response.status_code == 401:
            raise Exception(f"Client must be authenticated to access this resource:{response.json()} request:{j}")
        if response.status_code == 500:
            raise Exception(f"Something went wrong. Please try again later:\"{response.json()}\". If the error persists, contact our support team, func:{coords}\n request:{j}")
        
        if response.status_code == 400:
            raise Exception(f"Server returned \"Bad Request\":\"{response.json()}\"\nrequest:{j}" )
            
        if ((response.status_code != 200) and (response.status_code != 201)):
            raise Exception(f"{coords} Returned code {response.status_code}: {response.json()}\n request:{j}")
        
        return response.json()

    ''' findObjectSchemas: Returns a vector of all Object Schemas
        @returns a vector of dictionaries
    '''
    def findObjectSchemas(self):
        query = self.coords+"assets/"+self.version+"/objectschema/list"
        res = self._private_get(query)   
        return res

    ''' findObjectTypesFlatList: Returns a vector of all known ObjectTypes from the current schema
        @returns a vector of dictionaries
    '''
    def findObjectTypesFlatList(self):
        query = self.coords+"assets/"+self.version+"/objectschema/"+ self.schemaId+"/objecttypes/flat"
        res = self._private_get(query)   
        return res

    ''' getObjectTypeIdFromName: Returns the object type ID of the object type with a specified name
        @name the exact name of the object type
        @returns integer
    '''
    def getObjectTypeIdFromName(self,name):
        #this call for updating the list before searching in it.
        self.objectTypeList=self.findObjectTypesFlatList()
        for objt in self.objectTypeList:
            if objt['name']==name:
                return objt['id']
        raise Exception("No object present with name "+name + " in the current object schema")


    ''' getObjectTypeNameFromID: Returns the object type name of the object type with a specified name
        @name the exact name of the object type
        @returns a string
    '''
    def getObjectTypeNameFromID(self,id):
        #this call for updating the list before searching in it.
        self.objectTypeList=self.findObjectTypesFlatList()
        for objt in self.objectTypeList:
            if objt['id']==id:
                return objt['name']
        raise Exception("No object present with id "+id + " in the current object schema")


    ''' listKnownObjectTypes: Returns a list of the known object types in the database..
        @returns a list of strings.
    '''
    def listKnownObjectTypes(self):
        #this call for updating the list before searching in it.
        self.objectTypeList=self.findObjectTypesFlatList()
        ret=[] 
        for objt in self.objectTypeList:
            ret.append(objt['name'])
        return ret



    def FindObjectsAql(self,aqlquery):
        #/rest/assets/{version}/aql/objects
        #query = self.coords+"assets/"+self.version+"/aql/objects/{}".format("?resultPerPage=1&page=3")
        query = self.coords+"assets/"+self.version+"/aql/objects"
        query = self.coords+"assets/"+self.version+"/aql/objects/{}".format("?qlQuery="+aqlquery)
        res = self._private_get(query)   
        return res


    '''findStatusTypes returns a list with the status type for the current ObjectSchema:
    @returns a vector of dictionaries
    '''
    def findStatusTypes(self):
        
        query=self.coords+"assets/"+self.version+"/config/statustype"
        res = self._private_get(query,{"objectSchemaId":self.schemaId}) 
        return res

    '''storeStatusType creates a new status type:
    @name  name of the new status
    @category category of the new status
    '''
    def storeStatusType(self,name,category):
        newstatus={'name': name,'category':str(category),'objectSchemaId':self.schemaId}
        query = self.coords+"assets/"+self.version+"/config/statustype"

        return self._private_generic_post(query,newstatus)

    '''getStatusType returns the Status Type with the id passed as parameter
    @id the Status ID to found
    '''
    def getStatusType(self,id):
        query= self.coords+"assets/"+self.version+"/config/statustype/"+str(id)
        return self._private_get(query)
    
    '''deleteStatusType Deletes the Status Type with the id passed as parameter
    @id the Status ID to delete
    '''
    def deleteStatusType(self,id):
        query= self.coords+"assets/"+self.version+"/config/statustype/"+str(id)
        self._private_delete(query)
        
    def UpdateStatusType(self,id,name,category):
        newstatus={'name': name,'category':str(category),'objectSchemaId':self.schemaId}
        query= self.coords+"assets/"+self.version+"/config/statustype/"+str(id)
        self._private_put(query,newstatus)

    '''findGlobalIcons  
    Find all global icons i.e. icons not associated with a particular object schema.
    @returns a vector of dictionaries
    '''
    def findGlobalIcons(self):
        query=self.coords+"assets/"+self.version+"/icon/global"
        return self._private_get(query)

    '''findIcons  
    Find all  icons associated to an object schema.
    @returns a vector of dictionaries
    '''
    def findIcons(self):
        query=query=self.coords+"assets/"+self.version+"/icon/objectschema/"+str(self.schemaId)
        return self._private_get(query)

    '''getIcon returns the Icon with the id passed as parameter. The Icon could be part of any schemaId
    @id the Status ID to found
    generate 500 exception code if the ID doesn't exists
    '''
    def getIcon(self,id):
        query= self.coords+"assets/"+self.version+"/icon/"+str(id)
        return self._private_get(query)



    def searchObjects(self,aqlquery,objectType="",resultsPerPage=10000,page=1,includeAttributes=False):
        query = self.coords+"assets/"+self.version+"/object/navlist/aql"
        params={
            'objectSchemaId':self.schemaId,
            "qlQuery":aqlquery,
            "page": page,
		    "resultsPerPage": resultsPerPage,
 		    "includeAttributes": includeAttributes
           #"filters":[]
               }
        if objectType:
            if objectType == "":
                params["objectTypeId"]=""
            else:
                params["objectTypeId"]=self.mapNameType[objectType]['id']
        else:
            #query = self.coords+"assets/"+self.version+"/aql/objects"
            query = self.coords+"assets/"+self.version+f"/aql/objects"
            params={
                "qlQuery":aqlquery,
                "page": page,
		        "resultsPerPage": resultsPerPage,
 		        "includeAttributes": includeAttributes,
                "includeTypeAttributes":True,
                "includeExtendedInfo":True
                
            }
            return self._private_get(query,params)
     
        
        return self._private_generic_post(query,params)


    
    def _private_updateObject(self,jsonDescr,id):
        query = self.coords+"assets/"+self.version+"/object/"+str(id)

        ret = self._private_put(query,jsonDescr)
        return ret

    def _private_createObject(self,jsonDescr,id):
        if id:
            ret = self._private_updateObject(jsonDescr,id)
            return ret
            
        else:
            query = self.coords+"assets/"+self.version+"/object/create"
            ret = self._private_generic_post(query,jsonDescr)
            return ret
        
    '''deleteObject Deletes the object with the id passed as parameter
    @id the  ID of the object to delete
    '''
    def deleteObject(self,id):
        query = self.coords+"assets/"+self.version+"/object/"+str(id)
        self._private_delete(query)
    
    ''' getObject  returns a single object with the objectId passed as parameter, of the type objectTypeId
    @objectId the object ID to retrieve
    @objectTypeId  the type of the object to retrieve
    '''
    def getObject(self,objectId,objectTypeId):
        ans=self.searchObjects("objectId="+str(objectId),objectTypeId,includeAttributes=True)
        if ans['totalFilterCount'] == 0:
            raise Exception("object with id "+str(objectId)+" not found")
        if ans['totalFilterCount'] > 1:
            raise Exception("something went wrong.. Multiple objects with same Id")

        return ans['objectEntries'][0]


    ''' getObjectFromName  returns a single object with the name passed as parameter, of the type objectTypeId
    @name the object name to retrieve
    @objectTypeId  the type of the object to retrieve
    raise exception when multiple object have same Name..
    '''
    def getObjectFromName(self,name,objectTypeId):
        ans=self.searchObjects("Name="+name,objectTypeId,includeAttributes=True)
        if ans['totalFilterCount'] == 0:
            raise Exception("object with Name "+name+" not found")
        if ans['totalFilterCount'] > 1:
            raise Exception("something went wrong.. Multiple objects with same Name")

        return ans['objectEntries'][0]

    '''findObjectAttributes   returns a vector with all the attributes associated to the object with a particular id
    @id the object id 
    '''
    def findObjectAttributes(self,id):
        #GET /rest/assets/{version}/object/{id}/attributes
        query=self.coords+"assets/"+self.version+"/object/"+str(id)+"/attributes"
        return self._private_get(query)

    '''findObjectAttributes   returns a vector with all the attributes associated to the objecttype with a particular id
    @id the objecttype id 
    '''
    def findObjectTypeAttributes(self,objectTypeId):
      
        query=self.coords+"assets/"+self.version+"/objecttype/"+str(objectTypeId)+"/attributes"
        return self._private_get(query)

   

    def findKeyFromName(self,name,objtype):
        if objtype+"-"+name in self.referenceCache:
            return self.referenceCache[objtype+"-"+name]
        q="Name="+name
        aql=self.searchObjects(q,objtype)
        if len(aql['objectEntries'])>0:       
            k=aql['objectEntries'][0]['objectKey']
            self.referenceCache[objtype+"-"+name]=k
            return k
        return None


    def createObject(self,objtypeName,attrList,name,id=None,refid=False):
        #trovo l' ID del tipo.
        newObject={}
        newObject['attributes']=[]
        if objtypeName not in self.mapNameAttribs:
            self.mapNameAttribs[objtypeName]=self.findObjectTypeAttributes(self.mapNameType[objtypeName]['id'])

        attribs=self.mapNameAttribs[objtypeName]
        
        
        attrList['Name']=name

        newObject['objectTypeId']=str(self.mapNameType[objtypeName]['id'])

        for el in attrList:
            found=False 

            #print(el)
            for a in attribs:
                if a['name'] == el:
                    found=True
                    stype="uknown"
                    attr_val=[]
                   

                    if a['type']==0:
                        ## text
                        stype=a['defaultType']['name']
                    elif a['type']==1:
                        stype="Object"
                    elif a['type']==7:
                        stype="Status"
                        found=None
                        if attrList[el] =="":
                            ## no state defined,
                            continue
                        
                        for s in self.schemaStates:
                            if s['name'] == attrList[el]:
                                attrList[el]=s['id']
                                found=s['id']
                        if not(found):
                            raise Exception(f"\"{attrList[el]}\" Not a valid State please correct or add a new state on the schema")

                            
                    
                    if a['editable'] != True:
                        raise Exception(f"Cannot set attribute {el} ({stype}) for the object ")
                    attId=a['id']

                    attrToAppend={}
                    attrToAppend['objectTypeAttributeId']=attId                        
                    attrToAppend['objectAttributeValues']=[]
                    if isinstance(attrList[el], list):
                        attr_val=attrList[el]
                    else:
                        attr_val.append(attrList[el])
                    for va in attr_val:
                        vals={}
                        
                        if 'referenceType' in a:
                            if va =='':
                                ## no reference
                                continue
                            if refid:
                                rkey = va ## already a key
                            else:
                                rkey=self.findKeyFromName(f"\"{va}\"",f"{a['referenceObjectType']['name']}")
                            
                            if rkey:
                                vals['value']=rkey
                            else:
                                raise Exception(f"Cannot find attribute reference@ \"{attrList[el]}\" in the attribute \"{a['referenceType']['name']}\"\n\tPlease create the missing reference before import")


                        else:
                            vals['value']=va
                        attrToAppend['objectAttributeValues'].append(vals)

                    newObject['attributes'].append(attrToAppend)
            if found == False:
                raise Exception("Cannot find attribute \""+el + "\" ({stype}) in the object of type "+objtypeName)
        
        ans=self._private_createObject(newObject,id)
        return ans

        
      
   

    


